# Copyright (c) 2013 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
from __future__ import absolute_import

from pyschema.core import (
    PySchema, Record, dumps, loads, ispyschema, SchemaStore,
    disable_auto_register, enable_auto_register, no_auto_store,
    NO_DEFAULT
)
from pyschema.types import *
import warnings
warnings.simplefilter('default')
